from flask import request

from api.routes import *
from api.job import (JOBS_MIN_PAGE_SIZE, JOBS_MAX_PAGE_SIZE, WORKER_ID_PATTERN, get_jobs, get_workers,
                     delete_ready_failed_job, restart_failed_job, cancel_running_job, copy_finished_deleted_job)

from api.miscellaneous import *


@api_route("/jobs", ["GET"], allow_while_readonly=True)
@api_moderator_route()
def api_route_jobs():
    entries_per_page = api_request_get_query_int("entries_per_page", 50, JOBS_MIN_PAGE_SIZE, JOBS_MAX_PAGE_SIZE)
    page = api_request_get_query_int("page", 0)
    type = api_request_get_query_string("type", 100, ID_STRING_PATTERN_NO_LENGTH)
    worker_id = api_request_get_query_string("worker_id", 100, WORKER_ID_PATTERN)
    state = api_request_get_query_string("state", 100)
    
    if state is not None and state not in ["ready", "running", "finished", "failed", "deleted"]:
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER("URL.state", "Unknown state"))
    
    page_count, page = get_jobs(entries_per_page, page, type, worker_id, state)
    return {
        "page_count": page_count,
        "page": page,
        "workers": get_workers()
    }


# These routes are so strict regarding the current state to prevent accidental actions
# if a job's state changed after the jobs were loaded by the user

@api_route("/job/<int:id>", ["DELETE"])
@api_moderator_route(require_csrf_token=True)
def api_route_delete_job(id):
    check_client_int(id, "URL.id")
    if not delete_ready_failed_job(id):
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(
            "URL.id", "This job is not ready or failed"))
    return {}


@api_route("/job/<int:id>/action", ["POST"])
@api_moderator_route(require_csrf_token=True)
def api_route_job_action(id):
    check_client_int(id, "URL.id")
    action = get_client_json(request).get_string("action", 100)
    match action:
        case "restart":
            if not restart_failed_job(id):
                raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(
                    "URL.id", "This job has not failed"))
            return {}
        case "cancel":
            if not cancel_running_job(id):
                raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(
                    "URL.id", "This job is not running"))
            return {}
        case "copy":
            if not copy_finished_deleted_job(id):
                raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(
                    "URL.id", "This job is not finished or deleted"))
        case _:
            raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(
                "$.action", "Unknown action"))
    return {}
