from flask import request

from api.objects.object import ObjectClass
from api.routes import *
from api.objects import *
from api.authentication import check_csrf_token


def _db_execute_path_configuration(trans: WriteTransaction,
                                   object_class: ObjectClass,
                                   modifying_user_id: int,
                                   object_id: int,
                                   expected_current_values_json: CJsonObject,
                                   updates_json: CJsonObject):
    object_class.modify_current_config(trans, modifying_user_id, object_id, expected_current_values_json, updates_json)
    trans.commit()


@api_route(f"""\
/object_management\
/<any({','.join(key for key, obj in object_classes_by_id.items())}):object_type>\
/<int:object_id>/configuration\
""", ["GET", "PATCH"], allow_while_readonly=True)
@api_moderator_route()
def api_route_object_management_configuration(object_type: str, object_id: int):
    if request.method == "PATCH":
        if api.live_config.is_readonly():
            raise ApiClientException(ERROR_SITE_IS_READONLY)
        check_csrf_token()
        check_client_int(object_id, "URL.object_id")
        json_request = get_client_json(request)
        db_pool.execute_write_transaction(_db_execute_path_configuration,
                                          object_classes_by_id[object_type],
                                          get_user_id(),
                                          object_id,
                                          json_request.get_object("expected_current_values"),
                                          json_request.get_object("updates"))
        return {}
    else:
        object_class = object_classes_by_id[object_type]
        if not object_class.is_get_config_allowed():
            raise ApiClientException(ERROR_UNKNOWN_REQUEST_PATH)
        return object_class.get_current_config(object_id)


@api_route(f"""\
/object_management\
/<any({','.join(key for key, obj in object_classes_by_id.items() if obj.is_creation_allowed())}):object_type>\
/new/configuration\
""", ["GET"], allow_while_readonly=True)
@api_moderator_route()
def api_route_object_management_new_configuration(object_type: str):
    return object_classes_by_id[object_type].get_creation_config()


@api_route(f"""\
/object_management\
/<any({','.join(key for key, obj in object_classes_by_id.items() if obj.is_creation_allowed())}):object_type>\
/new\
""", ["PUT"])
@api_moderator_route(require_csrf_token=True)
def api_route_object_management_new(object_type: str):
    modifying_user_id: int = get_user_id()
    
    json_request = get_client_json(request)
    parent_type = json_request.get_string("parent_type", 100, optional=True)
    parent_id = json_request.get_int("parent_id", MIN_VALUE_SINT32, MAX_VALUE_SINT32, optional=parent_type is None)
    if parent_id is not None and parent_type is None:
        raise ApiClientException(ERROR_REQUEST_MISSING_PARAMETER("parent_type"))
    values_json = json_request.get_object("values")
    variant = json_request.get("variant", optional=True)
    return {
        "id": object_classes_by_id[object_type].create(modifying_user_id, parent_type, parent_id, values_json, variant)
    }, HTTP_201_CREATED


_MAX_OBJECT_TYPE_LENGTH = max(100, max(map(lambda key: len(key), object_classes_by_id.keys())))


def _db_execute_modify_many(trans: WriteTransaction,
                            modifying_user_id: int,
                            updates: list[tuple[ObjectClass, int, CJsonObject, CJsonObject]]):
    for object_class, object_id, expected_current_values_json, updates_json in updates:
        object_class.modify_current_config(
            trans, modifying_user_id, object_id, expected_current_values_json, updates_json)
    trans.commit()


@api_route("/object_management/modify_many", ["POST"])
@api_moderator_route(require_csrf_token=True)
def api_route_object_management_modify_many():
    modifying_user_id: int = get_user_id()
    
    json_request = get_client_json(request)
    objects_list_json = json_request.get_array("objects")
    if objects_list_json.length() > 32:
        raise ApiClientException(ERROR_OBJECT_ERROR("Too many updates. Only 32 updates in one query are allowed"))
    
    seen_objects: dict[tuple[str, int], None] = {}
    updates: list[tuple[ObjectClass, int, CJsonObject, CJsonObject]] = []
    for object_json_raw in objects_list_json:
        object_json = object_json_raw.as_object()
        object_type = object_json.get_string("type", _MAX_OBJECT_TYPE_LENGTH)
        object_id = object_json.get_int("id", MIN_VALUE_SINT32, MAX_VALUE_SINT32)
        
        if object_type not in object_classes_by_id:
            raise ApiClientException(ERROR_OBJECT_ERROR(f"Unknown type {truncate_string(object_type)}"))
        if (object_type, object_id) in seen_objects:
            raise ApiClientException(ERROR_OBJECT_ERROR(
                f"Duplicated object of type {truncate_string(object_type)} with id {object_id}"))
        seen_objects[(object_type, object_id)] = None
        
        updates.append((
            object_classes_by_id[object_type],
            object_id,
            object_json.get_object("expected_current_values"),
            object_json.get_object("updates")
        ))
    db_pool.execute_write_transaction(_db_execute_modify_many, modifying_user_id, updates)
    return {}


@api_route(f"""\
/object_management\
/<any({','.join(key for key, obj in object_classes_by_id.items() if obj.is_deletion_allowed())}):object_type>\
/<int:object_id>\
""", ["DELETE"])
@api_moderator_route(require_csrf_token=True)
def api_route_object_management_delete(object_type: str, object_id: int):
    check_client_int(object_id, "URL.object_id")
    modifying_user_id: int = get_user_id()
    object_classes_by_id[object_type].delete(modifying_user_id, object_id)
    return {}


@api_route(f"""\
/object_management\
/<any({','.join(key for key, obj in object_classes_by_id.items() if obj.is_deletion_allowed())}):object_type>\
/<int:object_id>\
/resurrect\
""", ["POST"])
@api_moderator_route(require_csrf_token=True)
def api_route_object_management_resurrect(object_type: str, object_id: int):
    check_client_int(object_id, "URL.object_id")
    modifying_user_id: int = get_user_id()
    object_classes_by_id[object_type].undo_deletion(modifying_user_id, object_id)
    return {}


@api_route("/object_management/changelog", ["GET"], allow_while_readonly=True)
@api_moderator_route()
def api_route_object_management_changelog():
    entries_per_page = api_request_get_query_int("entries_per_page", 50, CHANGELOG_MIN_PAGE_SIZE, CHANGELOG_MAX_PAGE_SIZE)
    page = api_request_get_query_int("page", 0)
    object_type: str or None = api_request_get_query_string("object_type", 100, ID_STRING_PATTERN_NO_LENGTH, None)
    object_id: int or None = api_request_get_query_int("object_id", None)
    field_id: str or None = api_request_get_query_string("field", 100, ID_STRING_PATTERN_NO_LENGTH, None)
    
    object_class = None
    if object_type is not None:
        object_class = object_classes_by_id.get(object_type, None)
        if object_class is None:
            raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(
                "URL.object_type", f"Unknown object type {truncate_string(object_type)}"))
    
    subject = None
    if field_id is not None:
        if object_class is None:
            raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(
                "URL.object_type", "Must be present if 'field' is present"))
        if field_id not in object_class.get_all_field_ids():
            raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(
                "URL.object_type",
                f"Unknown field {truncate_string(field_id)} for object type {object_type}"))
        subject = object_class.get_field(field_id)
    
    page_count, log = get_changelog(
        entries_per_page,
        page,
        object_class,
        object_id,
        subject
    )
    return {
        "page_count": page_count,
        "page": log
    }
